
/***************************************************************************/
/***************************************************************************/
/**                                                                       **/
/**  Header Files                                                         **/
/**                                                                       **/
/***************************************************************************/
/***************************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include<Windows.h>



/***************************************************************************/
/***************************************************************************/
/**                                                                       **/
/**  Constant Definitions                                                 **/
/**                                                                       **/
/***************************************************************************/
/***************************************************************************/

/* 
*  Options included in the encoder
*/
#define OPTION_NEARLY_LOSSLESS 
#define OPTION_EXTRA_MODELING
#define OPTION_PREDICTOR_SELECTION
#define NUM_CONTEXT  18    /* Number of contexts used by the model */
#define NUM_BCONTEXT 64

#define THRESHOLD ((255+tolerance)/mask)
/*
*  Misc. constants
*/

#define N_SEQ     NUM_CONTEXT     /* Number of coding histograms        */
#define N_BSEQ    NUM_BCONTEXT    /* Number of binary coding histograms */

#define UPPER     255             /* Largest possible pixel value       */
#define ERR_SPAN  (UPPER + 1)


/***************************************************************************/
/***************************************************************************/
/**                                                                       **/
/**  Macro Definitions                                                    **/
/**                                                                       **/
/***************************************************************************/
/***************************************************************************/

#define ABS(X)      (((X) < 0)   ? -(X) : (X))
#define MIN(X,Y)    (((X) < (Y)) ?  (X) : (Y))
#define MAX(X,Y)    (((X) < (Y)) ?  (Y) : (X))



/***************************************************************************/
/***************************************************************************/
/**                                                                       **/
/**  Variable Definitions                                                 **/
/**                                                                       **/
/***************************************************************************/
/***************************************************************************/

int tolerance;
int mask;
int recon;
int temp_err;

static char exeFullPath[200]={0};
double MSE = 0.0;

int MaxError = 0;

static FILE *pre_err[16];
static char error_filename[20], file_indx[2];
static int num_indx;
static int num_cls = 16;

/***************************************************************************/
/**                                                                       **/
/**  Function : writeToFile                                                  **/
/**                                                                       **/
/***************************************************************************/
int  writeToFile( int k,int val)
{
	pre_err[0];
	if (k == 0)      {fprintf(pre_err[0], "%d\n", val);  }
	else if (k == 1) {fprintf(pre_err[1], "%d\n", val);  }
	else if (k == 2) {fprintf(pre_err[2], "%d\n", val); }
	else if (k == 3) {fprintf(pre_err[3], "%d\n", val);  }
	else if (k == 4) {fprintf(pre_err[4], "%d\n",val); }
	else if (k == 5) {fprintf(pre_err[5], "%d\n", val);  }
	else if (k == 6) {fprintf(pre_err[6], "%d\n", val); }
	else if (k == 7) {fprintf(pre_err[7], "%d\n", val);  }
	else if (k == 8) {fprintf(pre_err[8], "%d\n", val); }
	else if (k == 9) {fprintf(pre_err[9], "%d\n", val);  }
	else if (k == 10) {fprintf(pre_err[10], "%d\n", val); }
	else if (k == 11) {fprintf(pre_err[11], "%d\n", val);  }
	else if (k == 12) {fprintf(pre_err[12], "%d\n", val); }
	else if (k == 13) {fprintf(pre_err[13], "%d\n", val); }
	else if (k == 14) {fprintf(pre_err[14], "%d\n", val); }
	else if (k == 15) {fprintf(pre_err[15], "%d\n", val); }
	else{ sprintf(stderr,"invalid k is found\n");}
	return 0;
}


/***************************************************************************/
/**                                                                       **/
/**  Function : bentropy                                                  **/
/**                                                                       **/
/***************************************************************************/

double bentropy (dist, nptn)
	int *dist, *nptn;
{
	register int i, total;
	register double p, entropy;

	total = 0;

	for (i = 0; i < 3; ++i) total += dist[i];

	*nptn = total;

	entropy = 0.0;

	for (i = 0; i < 3; ++i) 
		if (dist[i] != 0)
		{
			p = (double) dist[i]/total;
			entropy += p*log (p);
		}

		return (-1.0*entropy/log (2.0));
}


/***************************************************************************/
/**                                                                       **/
/**  Function : entropy                                                   **/
/**                                                                       **/
/***************************************************************************/

double entropy (dist, nptn)
	int *dist, *nptn;
{
	register int i, total;
	register double p, entropy;

	total = 0;

	for (i = 0; i < ERR_SPAN; ++i) total += dist[i];

	*nptn = total;

	entropy = 0.0;

	for (i = 0; i < ERR_SPAN; ++i) 
		if (dist[i] != 0)
		{
			p = (double) dist[i]/total;
			entropy += p*log (p);
		}

		return (-1.0*entropy/log (2.0));
}


/***************************************************************************/
/**                                                                       **/
/**  Function : nh                                                        **/
/**                                                                       **/
/***************************************************************************/

static int T0[] =   {7,  2,   3,   2,   4,   6,   5,   4,   2};
static int T1[] =   {17, 6,   6,   5,   13,  62,  18,  18,  13};
static int T2[] =   {28, 11,  15,  21,  39,  89,  56,  53,  47};
static int T3[] =   {46, 23,  30,  45,  68,  124, 98,  89,  100};
static int T4[] =   {65, 47,  53,  67,  94,  172, 138, 124, 140};
static int T5[] =   {91, 72,  81,  116, 127, 230, 184, 183, 188} ;
static int T6[] =   {148,140, 159, 300, 165, 300, 219, 300, 300};


int nh (k,tolerance)
	int k;
{

	if (tolerance > 8) tolerance = 8;

	if (k <= T0[tolerance])      return (0);
	else if (k <= T1[tolerance]) return (1);
	else if (k <= T2[tolerance]) return (2);
	else if (k <= T3[tolerance]) return (3);
	else if (k <= T4[tolerance]) return (4);
	else if (k <= T5[tolerance]) return (5);
	else if (k <= T6[tolerance]) return (6);
	else return (7);
}


/***************************************************************************/
/**                                                                       **/
/**  Function : nremap                                                    **/
/**                                                                       **/
/***************************************************************************/

int nremap (err, p)
	int err, p;
{
	register int aerr;

	if (err == 0) return (0);

	if (err > 0) err = (err + tolerance)/mask;
	else         err = (err - tolerance)/mask;

	if (err == 0) return (0);

	aerr = ABS (err);

	if (p <= (UPPER/2))
	{
		if ((mask*aerr) <= (p + tolerance))
			if (err < 0) err = -(err << 1);
			else         err =  (err << 1) - 1;
		else
			err = (p + tolerance)/mask + aerr;
	}
	else
	{
		if ((mask*aerr) <= (UPPER - p + tolerance))
			if (err < 0) err = -(err << 1);
			else         err =  (err << 1) - 1;
		else
			err = (UPPER - p + tolerance)/mask + aerr;
	}

	return (err);
}


/***************************************************************************/
/**                                                                       **/
/**  Function : quant2                                                    **/
/**                                                                       **/
/**  Mapping to quantized label and then to reconstructed value           **/
/**                                                                       **/
/***************************************************************************/

int quant2 (diff)
	int diff;
{
	if (diff == 0) return (diff);

	if (diff > 0) diff = (diff + tolerance)/mask;
	else          diff = (diff - tolerance)/mask;

	diff = diff*mask;

	return (diff);
}


/***************************************************************************/
/***************************************************************************/
/**                                                                       **/
/**  Operations for loading images                                        **/
/**                                                                       **/
/***************************************************************************/
/***************************************************************************/

FILE *infile;


/***************************************************************************/
/**                                                                       **/
/**  Function : create_buffer                                             **/
/**                                                                       **/
/***************************************************************************/

void create_buffer (u0, u1, u2, width, filename)
	unsigned char **u0;
unsigned char **u1;
unsigned char **u2;
int width;
char *filename;
{
	char pre_err_file[200];
	/*
	*  Open the input file
	*/
	if (!(infile = fopen (filename, "rb")))
	{
		fprintf (stderr, "ERROR : cannot open input file %s!\n", filename);
		exit    (1);
	}
	else {
		fprintf (stderr, ">>Input: %s\n", filename);
	}
	//end
	/*
	*  Load the first line of the image
	*/

	if ((*u2 = (unsigned char *) calloc (width + 2, sizeof (unsigned char))) == NULL)
	{
		fprintf (stderr, "ERROR : not enough space for buffer!\n");
		exit    (1);
	}

	if ((fread (*u2 + 1, sizeof (unsigned char), width, infile)) != width)
	{
		fprintf (stderr, "ERROR : in reading image!\n");
		exit    (1);
	}

	(*u2)[0]       = (*u2)[1];        /* Duplicate the first pixel */
	(*u2)[width+1] = (*u2)[width];    /* Duplicate the last  pixel */

	/*
	*  Load the second line of the image
	*/

	if ((*u1 = (unsigned char *) calloc (width + 2, sizeof (unsigned char))) == NULL)
	{
		fprintf (stderr, "ERROR : not enough space for buffer!\n");
		exit    (1);
	}

	if ((fread (*u1 + 1, sizeof (unsigned char), width, infile)) != width)
	{
		fprintf (stderr, "ERROR : in reading image!\n");
		exit    (1);
	}

	(*u1)[0]       = (*u1)[1];        /* Duplicate the first pixel */
	(*u1)[width+1] = (*u1)[width];    /* Duplicate the last  pixel */

	/*
	*  Load the third line of the image
	*/

	if ((*u0 = (unsigned char *) calloc (width + 2, sizeof (unsigned char))) == NULL)
	{
		fprintf (stderr, "ERROR : not enough space for buffer!\n");
		exit    (1);
	}

	if ((fread (*u0 + 1, sizeof (unsigned char), width, infile)) != width)
	{
		fprintf (stderr, "ERROR : in reading image!\n");
		exit    (1);
	}

	(*u0)[0]       = (*u0)[1];        /* Duplicate the first pixel */
	(*u0)[width+1] = (*u0)[width];    /* Duplicate the last  pixel */

	// open the file to store prediction error
	strcpy(pre_err_file,exeFullPath);
	for (num_indx = 1; num_indx <= num_cls; num_indx++)
	{
		strcpy(error_filename,pre_err_file);
		strcat(error_filename, "pre_err");
		sprintf(file_indx,"%d",num_indx);
		strcat(error_filename, file_indx);
		strcat(error_filename, ".txt");
		if (!(pre_err[num_indx-1] = fopen (error_filename, "w+")))
		{
			fprintf (stderr, "ERROR : cannot open pre_err%d.txt\n",num_indx);
			exit    (1);
		}
		else {
			fprintf (stderr, ">> write prediction error into pre_err%d.txt \n",num_indx);
		}
	}
}


/***************************************************************************/
/**                                                                       **/
/**  Function : update_buffer                                             **/
/**                                                                       **/
/***************************************************************************/

void update_buffer (u0, u1, u2, width)
	unsigned char **u0;
unsigned char **u1;
unsigned char **u2;
int width;
{
	unsigned char *temp;

	temp = *u2;
	*u2  = *u1;
	*u1  = *u0;
	*u0  = temp;

	/*
	*  Load a new line from the image
	*/

	if ((fread (*u0 + 1, sizeof (unsigned char), width, infile)) != width)
	{
		fprintf (stderr, "ERROR : in reading image!\n");
		exit    (1);
	}

	(*u0)[0]       = (*u0)[1];        /* Duplicate the first pixel */
	(*u0)[width+1] = (*u0)[width];    /* Duplicate the last  pixel */
}


/***************************************************************************/
/**                                                                       **/
/**  Function : destory_buffer                                            **/
/**                                                                       **/
/***************************************************************************/

void destory_buffer (u0, u1, u2)
	unsigned char **u0;
unsigned char **u1;
unsigned char **u2;
{
	free (*u0);  *u0 = NULL;
	free (*u1);  *u1 = NULL;
	free (*u2);  *u2 = NULL;

	/*
	*  Close the input file
	*/
	fclose (infile);
	for (num_indx = 1; num_indx<=num_cls; num_indx++)
	{
		fclose (pre_err[num_indx-1]);
	}	

}


/***************************************************************************/
/***************************************************************************/
/**                                                                       **/
/**  Operations for compression                                           **/
/**                                                                       **/
/***************************************************************************/
/***************************************************************************/

int Th[8] = { 1, 3, 6, 10, 15, 21, 28, 36 };


/***************************************************************************/
/**                                                                       **/
/**  Function : encode_buffer                                             **/
/**                                                                       **/
/***************************************************************************/

void encode_buffer (u0, u1, u2, width, height)
	unsigned char *u0;
unsigned char *u1;
unsigned char *u2;
int width, height;
{
	register unsigned short ptn, ptn0, ptn1, ptn2, ptn3;

	int *N1, *N2, *N3; 
	int *S1, *S2, *S3;

	int *Dh, *Dv;
	int *E0, err, aerr;

	int H[N_SEQ][ERR_SPAN], BH[N_BSEQ][3];

	register int dh, dv, max, min, adjust, temp;
	register int symb1, symb2, bflag;

	register int nnww, nnw, nn, nne, nww, nw, n, ne, nee, www, ww, w, curr;
	register int x, i, j, k, total;

	register double sum;



	/** MODELING ERROR *********************************************************/

#ifdef OPTION_EXTRA_MODELING

	register unsigned short ptnC;

	int *NC;
	int *SC;

#endif

	/** MODELING ERROR *********************************************************/



	/** PREDICTOR SELECTION ****************************************************/

#ifdef OPTION_PREDICTOR_SELECTION

	register int Ts;
	register int Td;

	int Ns[N_SEQ+1];
	int Nd[N_SEQ+1];

#endif

	/** PREDICTOR SELECTION ****************************************************/

	/*
	*  Allocate space ("Gradient")
	*/

	if ((Dh = (int *) calloc (width + 2, sizeof (int))) == NULL)
	{
		fprintf (stderr, "ERROR : no enough space for Dh\n");
		exit    (1);
	}

	if ((Dv = (int *) calloc (width + 2, sizeof (int))) == NULL)
	{
		fprintf (stderr, "ERROR : no enough space for Dv\n");
		exit    (1);
	}

	memset (Dh, 0, (width + 2)*sizeof (int));
	memset (Dv, 0, (width + 2)*sizeof (int));

	/*
	*  Allocate space ("Prediction Error")
	*/

	if ((E0 = (int *) calloc (width + 2, sizeof (int))) == NULL)
	{
		fprintf (stderr, "ERROR : no enough space for E0\n");
		exit    (1);
	}

	aerr = err = 0;

	memset (E0, 0, (width + 2)*sizeof (int));

	/*
	*  Allocate space ("Context Modeling")
	*/

	if ((N1 = (int *) calloc (4096, sizeof (int))) == NULL)
	{
		fprintf (stderr, "ERROR : no enough space for N1\n");
		exit    (1);
	}

	if ((S1 = (int *) calloc (4096, sizeof (int))) == NULL)
	{
		fprintf (stderr, "ERROR : no enough space for S1\n");
		exit    (1);
	}

	if ((N2 = (int *) calloc (4096, sizeof (int))) == NULL)
	{
		fprintf (stderr, "ERROR : no enough space for N2\n");
		exit    (1);
	}

	if ((S2 = (int *) calloc (4096, sizeof (int))) == NULL)
	{
		fprintf (stderr, "ERROR : no enough space for S2\n");
		exit    (1);
	}

	if ((N3 = (int *) calloc (4096, sizeof (int))) == NULL)
	{
		fprintf (stderr, "ERROR : no enough space for N3\n");
		exit    (1);
	}

	if ((S3 = (int *) calloc (4096, sizeof (int))) == NULL)
	{
		fprintf (stderr, "ERROR : no enough space for S3\n");
		exit    (1);
	}

	ptn = ptn0 = ptn1 = ptn2 = ptn3 = 0;

	memset (N1, 0, 4096*sizeof (int));
	memset (S1, 0, 4096*sizeof (int));
	memset (N2, 0, 4096*sizeof (int));
	memset (S2, 0, 4096*sizeof (int));
	memset (N3, 0, 4096*sizeof (int));
	memset (S3, 0, 4096*sizeof (int));

	/** MODELING ERROR *********************************************************/

#ifdef OPTION_EXTRA_MODELING
	{
		if ((NC = (int *) calloc (5*5*5*5*3*3, sizeof (int))) == NULL)
		{
			fprintf (stderr, "ERROR : no enough space for NC\n");
			exit    (1);
		}

		if ((SC = (int *) calloc (5*5*5*5*3*3, sizeof (int))) == NULL)
		{
			fprintf (stderr, "ERROR : no enough space for SC\n");
			exit    (1);
		}

		ptnC = 0;

		memset (NC, 0, 5*5*5*5*3*3*sizeof (int));
		memset (SC, 0, 5*5*5*5*3*3*sizeof (int));
	}
#endif

	/** MODELING ERROR *********************************************************/

	/*
	*  Initialize variables
	*/

	for (k = 0; k < N_SEQ; ++k) 
		for (j = 0; j < ERR_SPAN; ++j) H[k][j] = 0;

	for (k = 0; k < N_BSEQ; ++k) 
		for (j = 0; j < 3; ++j) BH[k][j] = 0;


	/** PREDICTOR SELECTION ****************************************************/

#ifdef OPTION_PREDICTOR_SELECTION
	{
		for (i = 0; i < N_SEQ+1; i++) Ns[i] = Nd[i] = 0;
	}
#endif

	/** PREDICTOR SELECTION ****************************************************/

	/*
	*  First pixel in the first row
	*/
	curr = u2[1];
	temp = UPPER/4;  //Upper = 255
	err  = curr - temp;
	temp_err = err;
	err   = quant2 (err);

	// Add
	recon = temp + err;

	if (recon < 0)     recon = 0;
	if (recon > UPPER) recon = UPPER;

	MSE += (curr - recon)*(curr - recon);

	if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

	u2[0] = u2[1] = curr = recon;

	aerr = ABS (err);

	// fprintf(pre_err4, "%d\n", nremap (err, temp));
	writeToFile(3,nremap (err, temp));

	++H[3][nremap (err, temp)];

	//encode_symbol (3, nremap (err, temp), tolerance);

	/*
	*  Second pixel in the first row
	*/

	w    = u2[1];
	curr = u2[2];

	temp = w;
	k = nh (aerr << 2, tolerance);

	err  = curr - temp;
	temp_err = err;
	err   = quant2 (err);

	// Add
	recon = temp + err;
	if (recon < 0)     recon = 0;
	if (recon > UPPER) recon = UPPER;

	MSE += (curr - recon)*(curr - recon);

	if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;
	u2[2] = curr = recon;
	aerr = ABS (err);
	writeToFile(k, nremap (err, temp));

	++H[k][nremap (err, temp)];

	//encode_symbol (k, nremap (err, temp), tolerance);

	/*
	*  Rest of the first row
	*/

	for (x = 3; x <= width; ++x)
	{
		ww   = u2[x-2];
		w    = u2[x-1];
		curr = u2[x];
		temp = w;

		if (temp > UPPER)  temp = UPPER; 
		else if (temp < 0) temp = 0;

		k = nh ((aerr + ABS (ww - w)) << 2, tolerance);

		err  = curr - temp;
		temp_err = err;
		err   = quant2 (err);

		// Add
		recon = temp + err;

		if (recon < 0)     recon = 0;
		if (recon > UPPER) recon = UPPER;

		MSE += (curr - recon)*(curr - recon);

		if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

		u2[x] = curr = recon;

		aerr = ABS (err);

		writeToFile(k, nremap (err, temp));

		++H[k][nremap (err, temp)];

	//	encode_symbol (k, nremap (err, temp), tolerance);
	}

	u2[width+1] = u2[width];

	/*
	*  Second row first pixel
	*/
	n    = u2[1];
	curr = u1[1];
	temp = n;
	err = curr - temp;
	temp_err = err;
	err   = quant2 (err);

	// Add
	recon = temp + err;

	if (recon < 0)     recon = 0;
	if (recon > UPPER) recon = UPPER;

	MSE += (curr - recon)*(curr - recon);

	if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

	u1[0] = u1[1] = curr = recon;

	E0[1] = err;


	aerr  = ABS (err);

	//fprintf(pre_err3, "%d\n", nremap (err, temp)); 
	writeToFile(2,nremap (err, temp));

	++H[2][nremap (err, temp)];

	//encode_symbol (2, nremap (err, temp), tolerance);

	/*
	*  Second row second pixel
	*/

	nw   = u2[1];
	n    = u2[2];
	ne   = u2[3];
	w    = u1[1];
	curr = u1[2];

	temp = (w + n + 1)/2 + (ne - nw)/4;

	if (temp > UPPER)  temp = UPPER; 
	else if (temp < 0) temp = 0;

	k = nh (aerr + ABS (n - nw) + ABS (n - ne) + 2*ABS (w - nw), tolerance);
	err = curr - temp;

	temp_err = err;
	err   = quant2 (err);

	// Add
	recon = temp + err;

	if (recon < 0)     recon = 0;
	if (recon > UPPER) recon = UPPER;

	MSE += (curr - recon)*(curr - recon);

	if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

	u1[2] = curr = recon;

	E0[2] = err;
	aerr  = ABS (err);

	writeToFile(k, nremap (err, temp));

	++H[k][nremap (err, temp)];

	//encode_symbol (k, nremap (err, temp), tolerance);

	Dh[2] = ABS (curr - w);
	Dv[2] = ABS (curr - n);

	/*
	*  2nd row
	*/

	for (x = 3; x <= width - 1; ++x)
	{
		nw   = u2[x-1];
		n    = u2[x];
		ne   = u2[x+1];
		ww   = u1[x-2];
		w    = u1[x-1];
		curr = u1[x];

		temp = (w + n + 1)/2 + (ne - nw)/4;

		if (temp > UPPER)  temp = UPPER; 
		else if (temp < 0) temp = 0;

		k = nh (aerr + ABS (ww - w) + ABS (n - nw) + ABS (n - ne) + 2*ABS (w - nw), tolerance);
		err = curr - temp;

		temp_err = err;
		err   = quant2 (err);

		// Add
		recon = temp + err;

		if (recon < 0)     recon = 0;
		if (recon > UPPER) recon = UPPER;

		MSE += (curr - recon)*(curr - recon);

		if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

		u1[x] = curr = recon;

		E0[x] = err;
		aerr  = ABS (err);

		writeToFile(k, nremap (err, temp));
		++H[k][nremap (err, temp)];

		//encode_symbol (k, nremap (err, temp), tolerance);

		Dh[x] = ABS (curr - w);
		Dv[x] = ABS (curr - n);
	}

	/*
	*  Second row last pixel
	*/

	nw   = u2[width-1];
	n    = u2[width];
	ww   = u1[width-2];
	w    = u1[width-1];
	curr = u1[width];

	temp = (w + n + 1)/2;

	if (temp > UPPER)  temp = UPPER; 
	else if (temp < 0) temp = 0;

	k = nh (aerr + 2*ABS (ww - w) + 2*ABS (w - nw), tolerance);
	err = curr - temp;

	temp_err = err;
	err   = quant2 (err);

	// Add

	recon = temp + err;
	if (recon < 0)     recon = 0;
	if (recon > UPPER) recon = UPPER;
	MSE += (curr - recon)*(curr - recon);

	if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

	u1[width] = curr = recon;

	E0[width] = err;
	aerr      = ABS (err);
	writeToFile(k, nremap (err, temp));
	++H[k][nremap (err, temp)];

	//encode_symbol (k, nremap (err, temp), tolerance);

	u1[width+1] = u1[width];

	Dh[width] = ABS (curr - w);
	Dv[width] = ABS (curr - n);

	/*
	*  From 3rd row process each row
	*/

	for (j = 2; j < height; ++j)
	{
		/*
		*  First pixel of each row
		*/
		nn   = u2[1];
		n    = u1[1];
		ne   = u1[2];
		curr = u0[1];

		temp = n + (ne - nn)/4;

		if (temp > UPPER)  temp = UPPER; 
		else if (temp < 0) temp = 0;

		k = nh (2*ABS (n - ne) + 2*ABS (n - nn), tolerance);
		err  = curr - temp;
		temp_err = err;
		err   = quant2 (err);

		// Add
		recon = temp + err;

		if (recon < 0)     recon = 0;
		if (recon > UPPER) recon = UPPER;

		MSE += (curr - recon)*(curr - recon);

		if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

		u0[0] = u0[1] = curr = recon;


		aerr = ABS (err);

		writeToFile(k, nremap (err, temp));

		++H[k][nremap (err, temp)];

		//encode_symbol (k, nremap (err, temp), tolerance);

		/*
		////////////	 *  Second pixel of each row!!!!Note 1
		*/

		nn   = u2[2];
		nne  = u2[3];
		nw   = u1[1];
		n    = u1[2];
		ne   = u1[3];
		w    = u0[1];
		curr = u0[2];

		temp = (w + n + 1)/2 + (ne - nw)/4;

		if (temp > UPPER)  temp = UPPER; 
		else if (temp < 0) temp = 0;

		k = nh (aerr + ABS (n - ne) + ABS (n - nw) + ABS (n - nn) + ABS (ne - nne) + ABS (w - nw), tolerance);

		E0[1] = err;
		err   = curr - temp;

		temp_err = err;
		err   = quant2 (err);

		recon = temp + err;

		if (recon < 0)     recon = 0;
		if (recon > UPPER) recon = UPPER;

		MSE += (curr - recon)*(curr - recon);

		if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

		u0[2] = curr = recon;

		aerr  = ABS (err);

		writeToFile(k, nremap (err, temp));
		++H[k][nremap (err, temp)];

		//encode_symbol (k, nremap (err, temp), tolerance);

		/*
		*  Rest of the row
		*/

		for (x = 3; x <= width - 1; ++x)
			//for(x=3; x<=3; x++)
		{
			nw   = u1[x-1];
			n    = u1[x];
			ne   = u1[x+1];
			ww   = u0[x-2];
			w    = u0[x-1];
			curr = u0[x];

			symb1 = symb2 = w;

			Dv[x-1] = ABS (symb1 - nw);
			Dh[x-1] = ABS (symb1 - ww);

			if (ne != symb1) symb2 = ne;
			else if (n  != symb1) symb2 = n;
			else if (nw != symb1) symb2 = nw;
			else if (ww != symb1) symb2 = ww;

			bflag = 0;

			if (symb1 == symb2 & bflag == 1) /* only one symbol */
			{
				bflag = 1;
				if ((curr != symb1) && (ABS (symb1 - curr) <= tolerance))
				{
					recon = symb1;

					MSE += (curr - recon)*(curr - recon);

					if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

					u0[x] = curr = recon;
				}

				if (curr == symb1)
				{ 
					++BH[0][0];

			//		bencode_symbol (0, 0);

					E0[x-1] = err;
					aerr    = err = 0;
				}
				else 
				{
					/* only TWO symbols: symb1 and esc */

					++BH[0][1];

				//	bencode_symbol (0, 1);

					temp = symb1;

					E0[x-1] = err;
					err     = curr - temp;
					temp_err = err;
					err   = quant2 (err);
					recon = temp + err;
					if (recon < 0)     recon = 0;
					if (recon > UPPER) recon = UPPER;

					MSE += (curr - recon)*(curr - recon);

					if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

					u0[x] = curr = recon;
					aerr    = ABS (err);
					++H[N_SEQ-2][nremap (err, temp)];

			//		encode_symbol (N_SEQ-2, nremap (err, temp), tolerance);
				}
			}
			else
			{
				nn  = u2[x];
				nne = u2[x+1];
				if ((nne != symb1) && (nne != symb2)) bflag = 0;
				else if ((nn != symb1) && (nn != symb2)) bflag = 0;
				else if ((ne != symb1) && (ne != symb2)) bflag = 0;
				else if ((n  != symb1) && (n  != symb2)) bflag = 0;
				else if ((nw != symb1) && (nw != symb2)) bflag = 0;
				else if ((ww != symb1) && (ww != symb2)) bflag = 0;
				else bflag = 0; /* two symbols */
			}

			if (bflag == 2)
			{
				if (ww == symb1) ptn = 1; else ptn = 0;

				if (nw == symb1) ptn |= 02;
				if (n  == symb1) ptn |= 04;
				if (ne == symb1) ptn |= 010;
				if (nn == symb1) ptn |= 020;

				if (symb1 < symb2) ptn |= 040;

				/* pass out symb1, symb2, esc */
				if ((curr != symb1) && (curr != symb2) && ((ABS (symb1 - curr) <= tolerance) || (ABS (symb2 - curr) <= tolerance)))
				{
					if (ABS (symb1 - curr) <= ABS (symb2 - curr))
					{
						recon = symb1;

						MSE += (curr - recon)*(curr - recon);

						if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

						u0[x] = curr = recon;
					}
					else
					{
						recon = symb2;

						MSE += (curr - recon)*(curr - recon);

						if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

						u0[x] = curr = recon;
					}
				}

				if (curr == symb1)
				{ 
					++BH[ptn][0];

			//		bencode_symbol (ptn, 0);

					E0[x-1] = err;
					aerr    = err = 0;
				}
				else if (curr == symb2)
				{ 
					++BH[ptn][1];

				//	bencode_symbol (ptn, 1);

					E0[x-1] = err;
					aerr    = err = 0;
				}
				else
				{
					++BH[ptn][2];

			//		bencode_symbol (ptn, 2);

					temp = symb1;

					E0[x-1] = err;
					err     = curr - temp;
					temp_err = err;
					err   = quant2 (err);
					recon = temp + err;
					if (recon < 0)     recon = 0;
					if (recon > UPPER) recon = UPPER;

					MSE += (curr - recon)*(curr - recon);

					if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

					u0[x] = curr = recon;
					aerr    = ABS (err);
					writeToFile(k, nremap (err, temp));
					++H[N_SEQ-1][nremap (err, temp)];

				//	encode_symbol (N_SEQ-1, nremap (err, temp), tolerance);
				}
			}

			if (bflag == 0)
			{
				nnww = u2[x-2];
				nnw  = u2[x-1];
				nww  = u1[x-2];
				nee  = u1[x+2];
				www  = u0[x-3];

				dv = Dv[x-1] + Dv[x] + Dv[x+1]/2;
				dh = Dh[x-1] + Dh[x] + Dh[x+1];

				k = nh (dv + dh + aerr + ABS (E0[x]) + (ABS (E0[x-1]) + ABS (E0[x+1]))/2, tolerance);

				/* This computation can be made incremental to reduce running time */

				min = max = w;

				if (n  < min)      min = n;
				else if (n  > max) max = n;
				if (nw < min)      min = nw;
				else if (nw > max) max = nw;
				if (ne < min)      min = ne;
				else if (ne > max) max = ne;

				/* The GAP predictor */

				symb1 = dv - dh;

				Ts = (w + n)/2 + (ne - nw)/4; 

				if (symb1 > 28)       Ts = (Ts + w)/2;
				else if (symb1 > 8)   Ts = (Ts + (Ts << 1) + w)/4;
				else if (symb1 < -32) Ts = (Ts + n)/2;
				else if (symb1 < -10) Ts = (Ts + (Ts << 1) + n)/4;

				/* The median predictor */

				if (w < n) { symb1 = w; symb2 = n; }
				else       { symb1 = n; symb2 = w; }

				if (nw >= symb2)      Td = symb1;
				else if (nw <= symb1) Td = symb2;
				else Td = w + n - nw;

				/* Select one of the two predictors to use */

				if (dv - dh > 100)       temp = w;
				else if (dv - dh < -100) temp = n;
				else
					temp = ((Ns[N_SEQ] >= Nd[N_SEQ]) ? Ts : Td);


				/** MODELING ERROR *********************************************************/

#ifdef OPTION_EXTRA_MODELING
				{
					register int F1, F2, F3;

					if (dv - dh > 100)       F1 = dh;
					else if (dv - dh < -100) F1 = dv;
					else
					{
						F1 = (dh + dv)/2;

						if (dv - dh > 28)       F1 = (3*dh +   dv)/4;
						else if (dv - dh >   8) F1 = (5*dh + 3*dv)/8;
						else if (dv - dh < -32) F1 = (  dh + 3*dv)/4;
						else if (dv - dh < -10) F1 = (3*dh + 5*dv)/8;
					}

					F2 = aerr + ABS (E0[x]) + (ABS (E0[x-1]) + ABS (E0[x+1]))/2;

					ptnC = 0;

					if (temp - ww > 4)       ptnC = 3*ptnC + 1;
					else if (temp - ww < -4) ptnC = 3*ptnC + 2;
					else ptnC = 3*ptnC;

					if (temp - nn > 4)       ptnC = 3*ptnC + 1;
					else if (temp - nn < -4) ptnC = 3*ptnC + 2;
					else ptnC = 3*ptnC;

					if (temp - nw > 12)       ptnC = 5*ptnC + 1;
					else if (temp - nw >   4) ptnC = 5*ptnC + 2;
					else if (temp - nw < -12) ptnC = 5*ptnC + 3;
					else if (temp - nw <  -4) ptnC = 5*ptnC + 4;
					else ptnC = 5*ptnC;

					if (temp - ne > 12)       ptnC = 5*ptnC + 1;
					else if (temp - ne >   4) ptnC = 5*ptnC + 2;
					else if (temp - ne < -12) ptnC = 5*ptnC + 3;
					else if (temp - ne <  -4) ptnC = 5*ptnC + 4;
					else ptnC = 5*ptnC;

					if (temp - w > 12)       ptnC = 5*ptnC + 1;
					else if (temp - w >   4) ptnC = 5*ptnC + 2;
					else if (temp - w < -12) ptnC = 5*ptnC + 3;
					else if (temp - w <  -4) ptnC = 5*ptnC + 4;
					else ptnC = 5*ptnC;

					if (temp - n > 12)       ptnC = 5*ptnC + 1;
					else if (temp - n >   4) ptnC = 5*ptnC + 2;
					else if (temp - n < -12) ptnC = 5*ptnC + 3;
					else if (temp - n <  -4) ptnC = 5*ptnC + 4;
					else ptnC = 5*ptnC;

					if (NC[ptnC] > 3) F3 = 5*SC[ptnC]/NC[ptnC];
					else
						F3 = F1;

					k = nh (F1 + F2 + F3, tolerance);
				}
#endif

				/** MODELING ERROR *********************************************************/

				if (w  < temp) ptn0  = 01; else ptn0 = 0;
				if (n  < temp) ptn0 |= 02;
				if (nw < temp) ptn0 |= 04;
				if (ne < temp) ptn0 |= 010;
				if (nn < temp) ptn0 |= 020;
				if (ww < temp) ptn0 |= 040;

				if (2*w-ww < temp) ptn0 |= 0100;
				if (2*n-nn < temp) ptn0 |= 0200;

				ptn0 |= ((k/2) << 8);

				if (nww < temp) ptn1 = ptn0 | 02000; else ptn1 = ptn0;
				if (nnw < temp) ptn1 |= 04000; 

				if (nee < temp) ptn2 = ptn0 | 02000; else ptn2 = ptn0;
				if (nne < temp) ptn2 |= 04000; 

				if (err   < 0) ptn3 = ptn0 | 02000; else ptn3 = ptn0;
				if (E0[x] < 0) ptn3 |= 04000;

				if (E0[x-1] < 0) ptn3 |= 0100; else ptn3 &= 077677;
				if (E0[x+1] < 0) ptn3 |= 0200; else ptn3 &= 077577;

				if (www < temp)         ptn3 |= 020; else ptn3 &= 077757;
				if (2*nw - nnww < temp) ptn3 |= 040; else ptn3 &= 077737;

				symb1 = symb2 = 0;

				if (N1[ptn1] > 3) { symb1 += N1[ptn1]; symb2 += S1[ptn1]; }
				if (N2[ptn2] > 3) { symb1 += N2[ptn2]; symb2 += S2[ptn2]; }
				if (N3[ptn3] > 3) { symb1 += N3[ptn3]; symb2 += S3[ptn3]; }

				adjust = (symb2/(symb1 + 1));

				temp += adjust;
				if (temp > max)      temp = max;
				else if (temp < min) temp = min;
				symb1  = S1[ptn1] + S2[ptn2] + S3[ptn3];
				symb1 += err + E0[x] + E0[x-1] + E0[x+1];

				symb2  = aerr + ABS (E0[x]) + (ABS (E0[x+1]) + ABS (E0[x-1]))/2;
				symb2 += ABS (adjust);
				symb2 += ABS (n - w)/8;

				E0[x-1] = err;
				err     = curr - temp;
				temp_err = err;
				err   = quant2 (err);
				//Note 
				recon = temp + err;

				if (recon < 0)     recon = 0;
				if (recon > UPPER) recon = UPPER;

				MSE += (curr - recon)*(curr - recon);

				if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

				u0[x] = curr = recon;
				aerr    = ABS (err);
				if (symb1 < 0)
				{
					if (symb2 > Th[k])
					{
						writeToFile(k+8, nremap (-err, temp));

						++H[k+8][nremap (-err, temp)];

				//		encode_symbol (k+8, nremap (-err, temp), tolerance);
					}
					else
					{ 

						writeToFile(k, nremap (-err, temp));

						++H[k][nremap (-err, temp)];

				//		encode_symbol (k, nremap (-err, temp), tolerance);
					}
				}
				else
				{
					if (symb2 > Th[k])
					{
						writeToFile(k+8, nremap (err, temp));
						++H[k+8][nremap (err, temp)];

				//		encode_symbol (k+8, nremap (err, temp), tolerance);
					}
					else
					{
						writeToFile(k, nremap (err, temp));
						++H[k][nremap (err, temp)];

			//			encode_symbol (k, nremap (err, temp), tolerance);
					}
				}

				if (ABS (err) < (16*((k >> 1) + 1)))
				{
					if (N1[ptn1] == 63)
					{
						if (ABS (S1[ptn1]) > 128)
						{
							S1[ptn1] = ((S1[ptn1] + err)/4);
							N1[ptn1] = 16;
						}
						else 
						{
							S1[ptn1] = ((S1[ptn1] + err)/2);
							N1[ptn1] = 32;
						}
					}
					else
					{
						S1[ptn1] += err;
						N1[ptn1]++;
					}

					if (N2[ptn2] == 63) 
					{
						if (ABS (S2[ptn2]) > 128)
						{
							S2[ptn2] = ((S2[ptn2] + err)/4);
							N2[ptn2] = 16;
						}
						else
						{
							S2[ptn2] = ((S2[ptn2] + err)/2);
							N2[ptn2] = 32;
						}
					}
					else
					{
						S2[ptn2] += err;
						N2[ptn2]++;
					}

					if (N3[ptn3] == 63)
					{
						if (ABS (S3[ptn3]) > 128)
						{
							S3[ptn3] = ((S3[ptn3] + err)/4);
							N3[ptn3] = 16;
						}
						else
						{
							S3[ptn3] = ((S3[ptn3] + err)/2);
							N3[ptn3] = 32;
						}
					}
					else
					{
						S3[ptn3] += err;
						N3[ptn3]++;
					}
				}



				/** MODELING ERROR *********************************************************/

#ifdef OPTION_EXTRA_MODELING
				{
					if (ABS (err) < (16*((k >> 1) + 1)))
					{
						if (NC[ptnC] == 63)
						{
							if (SC[ptnC] > 128)
							{
								SC[ptnC] = ((SC[ptnC] + ABS (err))/4);
								NC[ptnC] = 16;
							} 
							else
							{
								SC[ptnC] = ((SC[ptnC] + ABS (err))/2);
								NC[ptnC] = 32;
							} 
						} 
						else 
						{
							SC[ptnC] += ABS (err);
							NC[ptnC]++;
						} 
					}
				}
#endif

				/** MODELING ERROR *********************************************************/


				/** PREDICTOR SELECTION ****************************************************/

#ifdef OPTION_PREDICTOR_SELECTION
				{
					if (ABS (curr - Ts) < 3*k+1)
					{
						++Ns[k];
						++Ns[N_SEQ];
					}

					if (ABS (curr - Td) < 3*k+1)
					{
						++Nd[k];
						++Nd[N_SEQ];
					}
				}
#endif

				/** PREDICTOR SELECTION ****************************************************/

			}
		}

		/*
		*  Last pixel
		*/

		nn   = u2[width];
		nw   = u1[width-1];
		n    = u1[width];
		ww   = u0[width-2];
		w    = u0[width-1];
		curr = u0[width];

		Dh[width-1] = ABS (w - ww);
		Dv[width-1] = ABS (w - nw);

		temp = (w + n + 1)/2;

		if (temp > UPPER)  temp = UPPER; 
		else if (temp < 0) temp = 0;

		k = nh (aerr + ABS (n - nw) + ABS (n - nn) + ABS (w - ww) + ABS (w - nw), tolerance);

		E0[width-1] = err;
		err         = curr - temp;
		temp_err = err;
		err   = quant2 (err);

		// Add
		recon = temp + err;
		if (recon < 0)     recon = 0;
		if (recon > UPPER) recon = UPPER;

		MSE += (curr - recon)*(curr - recon);

		if (ABS (MaxError) < ABS (curr - recon)) MaxError = curr - recon;

		u0[width] = curr = recon;
		aerr        = ABS (err);
		writeToFile(k, nremap (err, temp));
		++H[k][nremap (err, temp)];

	//	encode_symbol (k, nremap (err, temp), tolerance);

		u0[width+1] = u0[width];
		Dh[width] = ABS (curr - w);
		Dv[width] = ABS (curr - n);

		E0[width] = err;

		/*
		*  Update input buffer here
		*/

		if (j < (height - 1)) update_buffer (&u0, &u1, &u2, width);
	}

	/*
	*  Compute the entropy
	*/

	sum = 0.0;
	total = 0;

	for (k = 0; k < N_SEQ; ++k)
	{
		double ep;
		int count;

		ep = entropy (H[k], &count);

		sum   += count*ep;
		total += count;
	}

	for (k = 0; k < N_BSEQ; ++k)
	{
		double ep;
		int count;

		ep = bentropy (BH[k], &count);

		sum   += count*ep;
		total += count;
	}

	sum /= width*height;

	/*
	*  Free the memory
	*/

	free (N1);  free (S1);
	free (N2);  free (S2);
	free (N3);  free (S3);



	/** MODELING ERROR *********************************************************/

#ifdef OPTION_EXTRA_MODELING
	{
		//ZN: begin
		//free (N3);  free (S3);
		//end
	}
#endif

	/** MODELING ERROR *********************************************************/
	free (Dh);  free (Dv);

	free (E0);

	/*
	*  Return the entropy
	*/

	fprintf (stderr, "Total Entropy = %lf\n", sum);
	 


	MSE /= width*height;

	//fprintf (stderr, "MSE = %lf, PSNR = %lf, Maximum error = %d\n", MSE, 10.0*log (256.0*256.0/MSE)/log (10.0), MaxError);
	if( MSE==0 ) {
		fprintf (stderr, "MSE = %lf, Maximum error = %d\n", MSE, abs(MaxError));
	}
	else {
		fprintf (stderr, "MSE = %lf, PSNR = %lf, Maximum error = %d\n", MSE, 10.0*log (255.0*255.0/MSE)/log (10.0), abs(MaxError));
	}

}


/***************************************************************************/
/***************************************************************************/
/**                                                                       **/
/**  Function : main                                                      **/
/**                                                                       **/
/***************************************************************************/
/***************************************************************************/
//ZN: begin
FILE *fStream;
//end

main (argc, argv)
	int argc;
char *argv[];
{
	unsigned char *line_curr;
	unsigned char *line_1_up;
	unsigned char *line_2_up;

	int width, height, depth;
	char img_attr_file[200] = {0};

	DWORD len = GetModuleFileName(NULL, exeFullPath, 200);
	len = WideCharToMultiByte(CP_ACP, 0, exeFullPath, len, img_attr_file, sizeof(img_attr_file)-1, NULL, NULL);
	img_attr_file[len] = '\0';
	strcpy( exeFullPath,img_attr_file );
	strrchr( exeFullPath, '\\')[0]= 0;
	
	/*
	*  Check if the number of arguments is correct
	*/

	if( argc!=6 )
	{
		fprintf (stderr, "USAGE : %s Input_filename Width Height Depth PAE \n", argv[0]);
		exit    (1);
	}

	
	/*
	*  Determine the dimension of the input file
	*/

	width  = atoi (argv[2]);
	height = atoi (argv[3]);

	if ((depth = atoi (argv[4])) != 8)
	{
		fprintf (stderr, "ERROR : the input must be an 8-bits image!\n");
		exit    (1);
	}

	//ZN: begin
	tolerance = atoi(argv[5]);
	mask = 2*tolerance + 1;
	strcat(exeFullPath,"\\result\\");strcpy(img_attr_file,exeFullPath);strcat(img_attr_file,"image_attr.txt");
	
		printf("current directory is: %s\n", exeFullPath);
	if( (fStream=fopen(img_attr_file,"w+"))==NULL ) {
		fprintf (stderr, "Cannot open image_attr.txt\n");
		exit    (1);
	}
	else {
		short st;
		unsigned char uct;
		fprintf (stderr, ">>write into image_attr.txt\n");
		fprintf (stderr, ">>Width:%d Height:%d PAE:%d\n",width,height,tolerance);
		st = (short)width;
		fwrite(&st,sizeof(short),1,fStream);
		st = (short)height;
		fwrite(&st,sizeof(short),1,fStream);
		uct = (unsigned char) tolerance;
		fwrite(&uct,1,1,fStream);
	}
	fclose(fStream);
	//end
	create_buffer  (&line_curr, &line_1_up, &line_2_up, width, argv[1]);

	encode_buffer  (line_curr, line_1_up, line_2_up, width, height);

	destory_buffer (&line_curr, &line_1_up, &line_2_up);

}

/***************************************************************************/

